/* onode_index.c
   -------------
   A casually demented B+Tree designed to stay balanced
   and be optimized for access from disk (and in-core cache)

   $Id: onode_index.c,v 1.15 2003/10/20 07:18:11 stewart Exp $

   (C)2003 Stewart Smith
   Distributed under the GNU Public License
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "testkit/types.h"

#include "disk.h"
#include "super_block.h"
#include "space_bitmap.h"
#include "onode_index.h"

extern u64 next_free;

u64 onode_index_new_id(struct fcfs_onode_index *index)
{
  
}

struct fcfs_onode_index* onode_index_read(struct fcfs_disk *disk)
{
  struct fcfs_onode_index *index;
  struct fcfs_disk_block *block;
  
  index = (struct fcfs_onode_index*)malloc(sizeof(struct fcfs_onode_index));
  if(index==NULL)
    {fprintf(stderr,"Cannot allocate onode index structure to read into\n");
    abort();}
  block = disk_getblock(disk,disk->sb->onode_index_blocknr);
  index->root_block = block;
  index->root = (struct fcfs_onode_index_node*)block->data;

  index->disk = disk;

  return index;
}

struct fcfs_onode_index* onode_index_new(struct fcfs_disk *disk)
{
  struct fcfs_onode_index *index;

  index = (struct fcfs_onode_index*)malloc(sizeof(struct fcfs_onode_index));
  if(index==NULL)
    {fprintf(stderr,"Cannot allocate onode index structure\n"); abort();}

  index->root = NULL;
  index->disk = disk;

  return index;
}

int onode_index_write_leaf(struct fcfs_onode_index *index, struct fcfs_onode_index_leaf *leaf)
{
  struct fcfs_disk_block *block;
  
  block = disk_getblock(index->disk,leaf->block);
#ifdef DEBUG_ONODE_INDEX
  fprintf(stderr,"--WRITE LEAF--\n");
#endif
  disk_writeblock(block);
  disk_freeblock(block);

  return 1;
}

struct fcfs_onode_index_leaf *onode_index_leaf_new(struct fcfs_onode_index *index, struct fcfs_onode_index_node *parent, int position)
{
  int i;
  struct fcfs_block_run *br;
  struct fcfs_onode_index_leaf *leaf;
  struct fcfs_disk_block *block;

  //  printf("onode_index_leaf_new\n");
  br = ag_allocate_block(index->disk,
			 BLOCK_AG(index->disk,parent->block),
			 parent->block%index->disk->sb->ag_blocksnr,
			 1);
  block = disk_newblock(index->disk,BR_SECTOR_T(index->disk,br));
  leaf = (struct fcfs_onode_index_leaf*)block->data;
			
#ifdef DEBUG_ONODE_INDEX
  fprintf(stderr,"LEAF: AG: %d, Start: %d, Len: %d\n",br->allocation_group,br->start,br->len);
#endif

  leaf->magic1 = FCFS_ONODE_INDEX_LEAF_MAGIC1;
  leaf->id = index->disk->sb->onindex_next_id++;
  leaf->block = BR_SECTOR_T(index->disk,br);  
  leaf->used = 0ULL;

  for(i=0;i<index->disk->sb->onindex_node_size;i++)
    { leaf->items[i].key = 0; leaf->items[i].onode_blocknr = 0; }

  onode_index_write_leaf(index,leaf);

  parent->items[position].node_blocknr = BR_SECTOR_T(index->disk,br);
  if(parent->used < position)
    parent->used = position;

  onode_index_write_node(index,parent);

  free(br);
  return leaf;
}

struct fcfs_onode_index_node* onode_index_read_node(struct fcfs_onode_index *index, u64 blocknr)
{
  struct fcfs_onode_index_node* node;
  struct fcfs_disk_block *block;
  
  block = disk_getblock(index->disk, blocknr);

  node = (struct fcfs_onode_index_node*) block->data;

  return node;
}

int onode_index_free_node(struct fcfs_onode_index *index, struct fcfs_onode_index_node* node)
{
  struct fcfs_disk_block *block;
  block = disk_getblock(index->disk,node->block);
  disk_freeblock(block);
  free(node);
  return 1;
}

int onode_index_write_node(struct fcfs_onode_index *index, struct fcfs_onode_index_node* node)
{
  struct fcfs_disk_block *block;

  /* Write root to disk */
  block = disk_getblock(index->disk,node->block);
  disk_writeblock(block);
#ifdef DEBUG_ONODE_INDEX
  fprintf(stderr,"--WRITE NODE--\n");
#endif
  disk_freeblock(block);
  
  return 1;
}

/* Writes root back to disk */
static inline int onode_index_write_root(struct fcfs_onode_index *index)
{
  return onode_index_write_node(index,index->root);
}

/* The ever belated onode_index_new_node */
struct fcfs_disk_block *onode_index_new_node(struct fcfs_onode_index *index)
{
  struct fcfs_block_run *br;
  struct fcfs_disk_block *block;
  struct fcfs_onode_index_node *node;
  int i;

  br = ag_allocate_block(index->disk,0,1,1);
  if((br->len) < 1)
    {
      fprintf(stderr,"Unable to allocate Onode Index Node\n");
      fprintf(stderr,"\tAG: %d, Start: %d, Len: %d\n",br->allocation_group,br->start,br->len);
      abort();
    }
  
  fprintf(stderr,"Onode Index node created at: AG: %d, Start: %d, Len: %d\n",br->allocation_group,br->start,br->len);

  block = disk_newblock(index->disk,BR_SECTOR_T(index->disk,br));
  node = (struct fcfs_onode_index_node*)block->data;
  
  node->magic1 = FCFS_ONODE_INDEX_NODE_MAGIC1;
  node->id = index->disk->sb->onindex_next_id++;
  node->used = 0ULL;
  node->block = br->start;

  for(i=0;i<index->disk->sb->onindex_node_size;i++)
    { node->items[i].key = 0; node->items[i].node_blocknr = 0;}

  return block;
}

/* Creates a new root node for an onode_index */
int onode_index_new_root(struct fcfs_onode_index *index)
{
  struct fcfs_onode_index_node *node;
  struct fcfs_disk_block *block;

  block = onode_index_new_node(index);
  index->root = (struct fcfs_onode_index_node*)block->data;
  index->root_block = block;

  fprintf(stderr,"****ONODE INDEX**** %llu\n",block->blocknr);

  /* Move br to index structure */
  /* Also copy into super block, and write to disk */
  /* update the SB records */
  index->disk->sb->onode_index_blocknr = block->blocknr;
  index->disk->sb->onindex_free = index->disk->sb->onindex_node_size;
  fcfs_write_sb(index->disk);
  onode_index_write_root(index);

  return 1;
}

int onode_index_leaf_insert(struct fcfs_disk *disk,struct fcfs_onode_index_leaf *leaf, u64 key, u64 value)
{
  int pos;
  int i;

  if(leaf->used==disk->sb->onindex_node_size)
    {				/* Need to split leaf */
      fprintf(stderr,"onode_index_leaf_insert: full onode_index_leaf. Growing of leaves not yet implemented\n");
      abort();
    }

  if(key > leaf->items[leaf->used-1].key)
    {				/* Append to end of leaf */
      leaf->used++;		/* FIXME: should be atomic */
      leaf->items[leaf->used-1].key = key;
      leaf->items[leaf->used-1].onode_blocknr = value;
      return 1;
    }
  else
    {
      for(i=0;i<leaf->used && leaf->items[i].key < key;i++)
	;			/* FIXME: Make binary search */
      pos = i;
      leaf->used++;
      for(i=leaf->used;i>pos;i--) /* Shuffle everything down */
	{
	  leaf->items[i].key = leaf->items[i-1].key;
	  leaf->items[i].onode_blocknr = leaf->items[i-1].onode_blocknr;
	}
      leaf->items[pos].key = key; /* Add our key */
      leaf->items[pos].onode_blocknr = value; /* Add our value */
      return 1;			/* Success! */
    }

  return 1;
}

/* Insert an onode into an index */
struct fcfs_block_run* onode_index_insert(struct fcfs_onode_index *index, struct fcfs_onode1 *onode)
{
  struct fcfs_block_run *onode_br;
  struct fcfs_disk_block *block;
  struct fcfs_onode_index_node *node;
  struct fcfs_onode_index_node *parent;
  struct fcfs_onode_index_leaf *leaf;
  int i,node_pos;

  if(index->root==NULL)
    onode_index_new_root(index);

  /* Determine ID for onode */
  //  onode->onode_num = ((~0ULL)/index->disk->sb->onindex_node_size) * (i+1);
  onode->onode_num = random();
  //index->disk->sb->onindex_next_id++;
  fcfs_write_sb(index->disk);

#ifdef DEBUG_ONODE_INDEX
  fprintf(stderr,"STORE ONODE ID: %llu\n",onode->onode_num);
#endif

  /* Revision 1 now that we're going onto disk */
  onode->onode_revision = 1;

  /* Use count = 1, in onode_index */
  onode->use_count = 1;

  /* Allocate room for onode */      
  //  printf("onode_index_insert\n");
  onode_br = ag_allocate_block(index->disk,0,10,1);

#ifdef DEBUG_ONODE_INDEX
  fprintf(stderr,"ONODE created at: AG: %d, Start: %d, Len: %d\n",onode_br->allocation_group,onode_br->start,onode_br->len);
#endif

  /* Write onode to disk */
  block = disk_newblock(index->disk,BR_SECTOR_T(index->disk,onode_br));
  //  memset(block->data,0,index->disk->bsize);
  memcpy(block->data,onode,sizeof(struct fcfs_onode1));
  disk_writeblock(block);
#ifdef DEBUG_ONODE_INDEX
  fprintf(stderr,"ONODE Written\n");
#endif
  disk_freeblock(block);

  /* Put it in the index properly */
  node = index->root;
   parent = NULL;

   /* Check current node for free space/location */
   for(i=0;i<node->used && node->items[i].key < onode->onode_num;i++)
     ;				/* FIXME: This should be a binary search */

   if(i==index->disk->sb->onindex_node_size)
     i = 0;

   node_pos = i;			/* position in the node */

#ifdef DEBUG_ONODE_INDEX
   fprintf(stderr,"\n\n\ti = %d\n\n",i);
#endif

   if(node->items[i].node_blocknr!=0)
     {
       /* Lookup subtree, see if node or leaf */
       parent = node;
       //      node = ;
       block = disk_getblock(index->disk,node->items[i].node_blocknr);
       if(*((u64*)block->data)== FCFS_ONODE_INDEX_LEAF_MAGIC1)
	 {			/* is a leaf */
#ifdef DEBUG_ONODE_INDEX
	   fprintf(stderr,"FOUND A LEAF!\n\n");
#endif
	   leaf = (struct fcfs_onode_index_leaf *)block->data;
#ifdef DEBUG_ONODE_INDEX
	   fprintf(stderr,"GOING TO ADD TO LEAF 0x%llu\n",leaf->id);
	   fprintf(stderr,"\tblock %llu\n",leaf->block);
	   fprintf(stderr,"\tused  %llu\n",leaf->used);
#endif

	   onode_index_leaf_insert(index->disk,leaf,
				   onode->onode_num,
				   BR_SECTOR_T(index->disk,
					       onode_br));

	   disk_writeblock(block);
	   parent->items[node_pos].key = onode->onode_num;
	   onode_index_write_node(index,parent);
	   index->disk->sb->onindex_used++;
	   index->disk->sb->onindex_free--;
	   fcfs_write_sb(index->disk);
	 }
       else if(*((u64*)block->data)== FCFS_ONODE_INDEX_NODE_MAGIC1)
	 {
	   fprintf(stderr,"FOUND A NODE!\n\n");
	   fprintf(stderr,"**FIXME** We don't handle nodes yet\n");
	 }
       else
	 {
	   fprintf(stderr,"WARNING-CORRUPT: node pointer points to corrupt node or leaf!\n");
	   fprintf(stderr,"\tCORRUPT node->items[%d].node_blocknr = %llu\n",i,node->items[i].node_blocknr);
	   fprintf(stderr,"\tMAGIC is %llx\n",*((u64*)block->data));
	   abort();
	 }
       disk_freeblock(block);
     }
   else
     {
       /* Create leaf */
       leaf = onode_index_leaf_new(index, node,i);
       leaf->items[0].key = onode->onode_num;
       leaf->items[0].onode_blocknr = BR_SECTOR_T(index->disk,onode_br);
       leaf->used = 1;

#ifdef DEBUG_ONODE_INDEX
       fprintf(stderr,"LEAF: Should be writing with key=%lld, block=%lld\n",leaf->items[0].key,leaf->items[0].onode_blocknr);
#endif

       node->items[i].node_blocknr = leaf->block;
       node->items[i].key = onode->onode_num;
       node->used = i+1;

       onode_index_write_root(index);      
       onode_index_write_leaf(index,leaf);
     }

   return onode_br;
 }

struct fcfs_disk_block *onode_index_lookup(struct fcfs_onode_index *index,struct fcfs_block_run *onode_br, u64 id)
{
  struct fcfs_disk_block *block;
  struct fcfs_disk_block *onode_block;
  struct fcfs_onode_index_node *node;
  struct fcfs_onode_index_node *parent;
  struct fcfs_onode_index_leaf *leaf;
  
  int i;
  
  node = index->root;
  
  for(i=0;i<node->used && node->items[i].key < id ;i++)
    ;

  /* Lookup subtree, see if node or leaf */
  parent = node;
  //      node = ;
  block = disk_getblock(index->disk,node->items[i].node_blocknr);
  if(*((u64*)block->data)== FCFS_ONODE_INDEX_LEAF_MAGIC1)
    {			/* is a leaf */
      fprintf(stderr,"FOUND A LEAF!\n\n");
      leaf = (struct fcfs_onode_index_leaf *)block->data;
      fprintf(stderr,"READ LEAF 0x%llu\n",leaf->id);
      fprintf(stderr,"\tblock %llu\n",leaf->block);
      fprintf(stderr,"\tused  %llu\n",leaf->used);
      for(i=0;i<leaf->used && leaf->items[i].key < id;i++)
	;
      fprintf(stderr,"leaf i = %d, %llu %llu\n",i,leaf->items[i].key,leaf->items[i].onode_blocknr);
      onode_block = disk_getblock(index->disk,leaf->items[i].onode_blocknr);
      onode_br->allocation_group = BLOCK_AG(index->disk,leaf->items[i].onode_blocknr);
      onode_br->start = leaf->items[i].onode_blocknr%index->disk->sb->ag_blocksnr;
      onode_br->len = 1;
      disk_freeblock(block);
    }
  else
    abort();
  
  return onode_block;
}

struct fcfs_onode_index* onode_index_delete(struct fcfs_onode_index* index)
{
  if(index->root)
    disk_freeblock(index->root_block);
  free(index);
  index = NULL;
  return index;
}
