/*
  block_dev.h
  -----------
  $Id: block_dev.h,v 1.13 2003/10/01 05:42:06 stewart Exp $
  
  Designed to be rather similar to the Linux Buffer Cache,
  i.e. with functions like bread() et al that function similarly.
  Should allow testing and development of filesystems purely in
  userspace and purely in an application.
  User Mode Linux could be used for this, but this suite is designed
  for even earlier in the design process.

  (C)2003 Stewart Smith
  Distributed under the GNU Public License.

  Some data structures have been constructed out of those
  present in the Linux Kernel (v2.5.69). They are copyright
  of their respective owners.
 */

#ifndef __BLOCK_DEV_H__
#define __BLOCK_DEV_H__

#include "types.h"
#include "bitops.h"
#include <glib.h>

/*
  struct block_device
  -------------------
  A really simple block_device structure.
  $Id: block_dev.h,v 1.13 2003/10/01 05:42:06 stewart Exp $
 */
struct block_device {
  char* name;
  int file_on_disk;
  u64 block_size;
  u64 num_blocks;

  /* Internal */
  u64 cache_hit;
  u64 cache_hit_clear;
  u64 cache_miss;
  u64 cache_miss_clear;
};

struct super_block {
  //struct list_head	s_list;		/* Keep this first */
  //dev_t			s_dev;		/* search index; _not_ kdev_t */
	unsigned long		s_blocksize;
	unsigned long		s_old_blocksize;
	unsigned char		s_blocksize_bits;
	unsigned char		s_dirt;
	unsigned long long	s_maxbytes;	/* Max file size */
  //struct file_system_type	*s_type;
  //struct super_operations	*s_op;
  //struct dquot_operations	*dq_op;
  //struct quotactl_ops	*s_qcop;
  //struct export_operations *s_export_op;
	unsigned long		s_flags;
	unsigned long		s_magic;
  //struct dentry		*s_root;
  //struct rw_semaphore	s_umount;
  //struct semaphore	s_lock;
	int			s_count;
	int			s_syncing;
	int			s_need_sync_fs;
  //atomic_t		s_active;
	void                    *s_security;

  //struct list_head	s_dirty;	/* dirty inodes */
  //struct list_head	s_io;		/* parked for writeback */
  //struct hlist_head	s_anon;		/* anonymous dentries for (nfs) exporting */
  //struct list_head	s_files;

	struct block_device	*s_bdev;
  //struct list_head	s_instances;
  //struct quota_info	s_dquot;	/* Diskquota specific options */

	char s_id[32];				/* Informational name */

  //struct kobject           kobj;          /* anchor for sysfs */
	void 			*s_fs_info;	/* Filesystem private info */

	/*
	 * The next field is for VFS *only*. No filesystems have any business
	 * even looking at it. You had been warned.
	 */
  //	struct semaphore s_vfs_rename_sem;	/* Kludge */
};

/*
  bh_state-bits
  -------------
  straight from include/linux/buffer_head.h (kernel 2.5.69)
 */
enum bh_state_bits {
  BH_Uptodate,    /* Contains valid data */
  BH_Dirty,       /* Is dirty */
  BH_Lock,        /* Is locked */
  BH_Req,         /* Has been submitted for I/O */

  BH_Mapped,      /* Has a disk mapping */
  BH_New,         /* Disk mapping was newly created by get_block */
  BH_Async_Read,  /* Is under end_buffer_async_read I/O */
  BH_Async_Write, /* Is under end_buffer_async_write I/O */
  BH_Delay,       /* Buffer is not yet allocated on disk */

  BH_Boundary,    /* Block is followed by a discontiguity */
  BH_PrivateStart,/* not a state bit, but the first bit available
		   * for private allocation by other entities
		   */
};

/*
  buffer_head
  -----------
  straight from include/linux/buffer_head.h (kernel 2.5.69)
  We've comment out things we don't really care too much about.
  $Id: block_dev.h,v 1.13 2003/10/01 05:42:06 stewart Exp $
 */
struct buffer_head {

  unsigned long b_state;          /* buffer state bitmap (see above) */
  atomic_t b_count;               /* users using this block */
  struct buffer_head *b_this_page;/* circular list of page's buffers */
  //  struct page *b_page;            /* the page this bh is mapped to */

  sector_t b_blocknr;             /* block number */
  u32 b_size;                     /* block size */
  char *b_data;                   /* pointer to data block */

  struct block_device *b_bdev;
  //  bh_end_io_t *b_end_io;          /* I/O completion */
  //  void *b_private;                /* reserved for b_end_io */
  //  struct list_head b_assoc_buffers; /* associated with another mapping */
};

/*
  Macro tricks to exapnd the set_buffer_foo(), clear_buffer_foo()
  and buffer_foo() functions
*/
#define BUFFER_FNS(bit,name) \
static inline void set_buffer_##name(struct buffer_head *bh)	\
{ \
  set_bit(BH_##bit, &(bh)->b_state);\
} \
static inline void clear_buffer_##name(struct buffer_head *bh) \
{ \
  clear_bit(BH_##bit, &(bh)->b_state); \
} \
static inline int buffer_##name(struct buffer_head *bh) \
{ \
  return test_bit(BH_##bit, &(bh)->b_state); \
}

/* 
   test_set_buffer_foo() and test_clear_buffer_foo()
*/
#define TAS_BUFFER_FNS(bit, name) \
static inline int test_set_buffer_##name(struct buffer_head *bh) \
{ \
  return test_and_set_bit(BH_##bit, &(bh)->b_state); \
} \
static inline int test_clear_buffer_##name(struct buffer_head *bh) \
{ \
  return test_and_clear_bit(BH_##bit, &(bh)->b_state); \
}

/*
  Emit the buffer bitops functions.
*/
BUFFER_FNS(Uptodate,uptodate)
BUFFER_FNS(Dirty,dirty)
TAS_BUFFER_FNS(Dirty,dirty)
BUFFER_FNS(Lock,locked)
TAS_BUFFER_FNS(Lock,locked)
BUFFER_FNS(Req,req)
BUFFER_FNS(Mapped,mapped)
BUFFER_FNS(New,new)
BUFFER_FNS(Async_Read,async_read)
BUFFER_FNS(Async_Write,async_write)
BUFFER_FNS(Delay,delay)
BUFFER_FNS(Boundary,boundary)



/*
  for submit_bh
*/
#define READ 1
#define WRITE 2

/*
  our function definitions. Mostly glue.
 */
int block_dev_new(struct block_device* b, const char* file, u64 block_size, u64 num_blocks);
int block_dev_close(struct block_device *b);

struct buffer_head *bnew(struct block_device *b, sector_t block, int size);

/*
  function definitions pretty much straight out of include/linux/block_head.h
 */
struct buffer_head *bread(struct block_device *b, sector_t block, int size);

/*
  reads block as per super block info
 */
static inline struct buffer_head *
sb_bread(struct super_block *sb, sector_t block)
{
	return bread(sb->s_bdev, block, sb->s_blocksize);
}

void block_dev_init();

void submit_bh(int rw,struct buffer_head * bh);

#endif
