/* volinfo.c
   ---------

   Dumps a lot of information about a volume to the console.

   $Id: volinfo.c,v 1.7 2003/09/22 09:05:18 stewart Exp $

   (C)2003 Stewart Smith
   Distributed under the GNU Public License
*/

// Block Device includes
#include "testkit/types.h"
#include "testkit/block_dev.h"
#include "testkit/bitops.h"
#include "disk.h"

// UNIX includes
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>

// FCFS Includes
#include "super_block.h"
#include "onode.h"
#include "onode_index.h"
#include "space_bitmap.h"

void print_sb_flags(struct fcfs_sb *sb, int desired_loc)
{
  if((sb->flags & 0x03)!=desired_loc)
    printf("WARNING-CORRUPT: ");
  if((sb->flags & 0x03) == FCFS_SBloc_start_volume)
    printf("This is the start of a volume\n");
  else
    if((sb->flags & 0x03) == FCFS_SBloc_end_volume)
      printf("This is the End of a volume\n");
    else
      if((sb->flags & 0x03) == FCFS_SBloc_start_ag)
	printf("This is the start of an Allocation Group\n");
      else
	printf("Volume is corrupt - unknown Super Block Location\n");

  printf("FLAGS: 0x%llx\n",sb->flags);
  
  if(test_bit(FCFS_FLAG_Dirty,&sb->flags))
    printf("\t- Volume is Dirty\n");
  if(test_bit(FCFS_FLAG_Experimental,&sb->flags))
    printf("\t- Volume has been used by EXPERIMENTAL code. It's probably b0rked.\n");
  if(test_bit(FCFS_FLAG_JournalMeta,&sb->flags))
    printf("\t- Volume journals Meta Data\n");
  if(test_bit(FCFS_FLAG_JournalData,&sb->flags))
    printf("\t- Volume journals Data\n");
  if(test_bit(FCFS_FLAG_Versioned,&sb->flags))
    printf("\t- Volume is Versioned\n");
}

int print_sb(struct fcfs_disk *disk)
{
  struct fcfs_disk_block *block, *block2;
  struct fcfs_sb *sb, *sb2;
  

  block = disk_getblock(disk,0);
  sb = (struct fcfs_sb*)block->data;
  if(sb->magic1!=FCFS_SB_MAGIC1)
    fprintf(stderr,"WARNING: Primary SB MAGIC1 mismatch. Corrupt Volume.\n");
  if(sb->magic2!=FCFS_SB_MAGIC2)
    fprintf(stderr,"WARNING: Primary SB MAGIC2 mismatch. Corrupt Volume.\n");
  if(sb->bits!=64)
    fprintf(stderr,"WARNING: Weird number of filesystem base bits. Corrupt Volume.\n");
  if(sb->version!=FCFS_SB_VERSION1)
    fprintf(stderr,"WARNING: Unknown Version of FS & SB.\n");

  printf("Primary Superblock\n");
  printf("------------------\n");
  printf("Volume name is: %s\n",sb->name);  /* FIXME: CHECK NAME!!! */
  printf("MAGIC1:  0x%x\tMAGIC2: 0x%x\nVERSION: 0x%x\tBITS: %u\t LENGTH: %u bytes\n",sb->magic1,sb->magic2,sb->version,sb->bits,sb->sblength);
  printf("Block Size: %u\tNo. Blocks: %llu\n",sb->bsize,sb->blocksnr);

  // sb flags
  print_sb_flags(sb,FCFS_SBloc_start_volume);

  printf("Number of Clean Mounts:   %llu\n",sb->num_mounts);
  printf("Number of UnClean Mounts: %llu\n",sb->num_dirtymounts);
  printf("Time Created:             %llu\n",sb->time_created);
  printf("Time last cleanly mounted:%llu\n",sb->time_clean);
  printf("Time last dirty mounted:  %llu\n",sb->time_dirtymount);

  printf("\n");
  
  printf("Onode Index:\n");
  printf("------------\n");
  printf("ONode Index location: %llu\n",sb->onode_index_blocknr);
  printf("Number of onodes: %llu used, %llu available, %llu total\n",
	 sb->onindex_used,
	 sb->onindex_free,
	 sb->onindex_num_onodes);
  printf("Next ID: %llu\tNode Size: %lu\n",
	 sb->onindex_next_id,
	 sb->onindex_node_size);
   

  block2 = disk_getblock(disk,disk->blocksnr-1);
  sb2 = (struct fcfs_sb*)block2->data;

  printf("\n\nChecking end of disk super block backup...\n");

  if(sb2->magic1!=sb->magic1)
    printf("WARNING-CORRUPT: End of disk SB magic1 is incorrect 0x%x\n",sb2->magic1);
  if(sb2->magic2!=sb->magic2)
    printf("WARNING-CORRUPT: End of disk SB magic2 is incorrect 0x%x\n",sb2->magic2);
  if(sb2->version!=sb->version)
    printf("WARNING-CORRUPT: End of disk SB version differs 0x%x\n",sb2->version);
  if(sb2->bits!=sb->bits)
    printf("WARNING-CORRUPT: End of disk SB bits differ %u\n",sb2->bits);
  if(sb2->sblength!=sb2->sblength)
    printf("WARNING-CORRUPT: End of disk SB length differs %u\n",sb2->sblength);

  print_sb_flags(sb2,FCFS_SBloc_end_volume);

  if((sb2->flags & (~0x03ULL))!=(sb->flags & (~0x03ULL)))
    {
      printf("WARNING-CORRUPT: End of disk SB flags differ\n");
      printf("End of Disk SB FLAGS: 0x%llx  0x%llx\n",sb2->flags, (sb2->flags & (~0x03ULL)));
    }

  if(sb2->bsize!=sb->bsize)
    printf("WARNING-CORRUPT: Block size differs %u vs %u\n",sb->bsize, sb2->bsize);
  if(sb2->blocksnr!=sb->blocksnr)
    printf("WARNING-CORRUPT: Number of blocks differs %llu vs %llu\n",sb->blocksnr,sb2->blocksnr);

  if(strcmp(sb2->name,sb->name)!=0)
    printf("WARNING-CORRUPT: Volume names differ.\n");
  
  disk_freeblock(block2);

  disk_freeblock(block);
}

int print_used_blocks(struct fcfs_disk *disk)
{
  int ag,blk;
  struct fcfs_disk_block *block;
  struct fcfs_sb *sb;
  
  block = disk_getblock(disk,0);
  sb = (struct fcfs_sb*)block->data;

  printf("Checking used blocks...\n");
  printf("-----------------------");

  for(ag=0;ag<sb->allocation_groupsnr;ag++)
    {
      printf("\nAG %u: ",ag);
      for(blk=0;blk<sb->ag_blocksnr;blk++)
	if(!ag_block_free(disk,ag, blk))
	  printf("%u,",blk);
    }
  printf("\n");
  return 0;
}

int print_onode_index_leaf(struct fcfs_disk *disk,u64 blocknr)
{
  struct fcfs_disk_block *block;
  struct fcfs_onode_index_leaf *leaf;
  int i;

  block = disk_getblock(disk,blocknr);
  leaf = (struct fcfs_onode_index_leaf *)block->data;

  if(leaf->magic1!=FCFS_ONODE_INDEX_LEAF_MAGIC1)
    printf("WARNING-CORRUPT: Bad onode_index_leaf MAGIC1\n");
  printf("\tOnode Index LEAF: %llu in block %llu\n",leaf->id,leaf->block);
  printf("\tUsed: %llu/%lu\n",leaf->used,disk->sb->onindex_node_size);
  printf("\t  key:block\n");
  for(i=0;i<leaf->used;i++)
    printf("\t  %llu:%llu\n",leaf->items[i].key,
	   leaf->items[i].onode_blocknr);  

  disk_freeblock(block);
  return 0;
}

int print_onode_index(struct fcfs_disk *disk)
{
  struct fcfs_disk_block *block;
  struct fcfs_onode_index_node *index_node;
  int i;

  block = disk_getblock(disk,disk->sb->onode_index_blocknr);
  
  index_node = (struct fcfs_onode_index_node*)block->data;
  if(index_node->magic1!=FCFS_ONODE_INDEX_NODE_MAGIC1)
    printf("WARNING-CORRUPT: Bad onode_index_node MAGIC1\n");
  printf("Onode Index Node: %llu in block %llu\n",index_node->id,index_node->block);
  if(disk->sb->onode_index_blocknr!=index_node->block)
    printf("WARNING-CORRUPT: onode_index_node->block != sb->onode_index_blocknr\n");
  printf("Used: %llu/%lu\n",index_node->used,disk->sb->onindex_node_size);
  printf("  key:block\n");
  for(i=0;i<index_node->used;i++)
    printf("  %llu:%llu\n",index_node->items[i].key,
	   index_node->items[i].node_blocknr);

  for(i=0;i<index_node->used;i++)
    print_onode_index_leaf(disk,index_node->items[i].node_blocknr);

  disk_freeblock(block);
  return 0;
}

int main(int argc, char* argv[])
{
  struct block_device bdev;
  struct fcfs_disk *disk;
  struct fcfs_disk_block *block;

  if(argc<4)
    {
      fprintf(stderr,"FCFS Volume Info Utility (volinfo)\n");
      fprintf(stderr,"----------------------------------\n");
      fprintf(stderr,"$Id: volinfo.c,v 1.7 2003/09/22 09:05:18 stewart Exp $\n\n");
      fprintf(stderr,"Written by: Stewart Smith (stewart@flamingspork.com)\n\n");
      fprintf(stderr,"Usage:\n\t./volinfo device blocksize blockcount\n\n");
      exit(0);
    }

  block_dev_init();
  block_dev_new(&bdev,argv[1],atoi(argv[2]),atoi(argv[3]));
  disk = disk_new(&bdev);

  block = disk_getblock(disk,0);
  disk->sb = (struct fcfs_sb*)block->data;

  print_sb(disk);

  print_used_blocks(disk);

  print_onode_index(disk);

  /* Clean up and exit */
  block_dev_close(&bdev);

  return 0;
}
